@@ -8,7 +8,7 @@
 This file is a part of bsc and/or libbsc, a program and a library for
 lossless, block-sorting data compression.
 
-   Copyright (c) 2009-2024 Ilya Grebnov <ilya.grebnov@gmail.com>
+   Copyright (c) 2009-2025 Ilya Grebnov <ilya.grebnov@gmail.com>
 
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
@@ -118,7 +118,7 @@ int bsc_get_cpu_features(void)
 
 int bsc_get_cpu_features(void)
 {
-    return LIBBSC_CPU_FEATURE_NONE;
+    return LIBBSC_CPU_FEATURE;
 }
 
 #endif
